@@ -48,6 +48,7 @@
 #include "upstream.h"
 #include "connect-ports.h"
 #include "conf.h"
+#include "basicauth.h"
 
 /*
  * Maximum length of a HTTP line
@@ -61,9 +62,11 @@
 #ifdef UPSTREAM_SUPPORT
 #  define UPSTREAM_CONFIGURED() (config.upstream_list != NULL)
 #  define UPSTREAM_HOST(host) upstream_get(host, config.upstream_list)
+#  define UPSTREAM_IS_HTTP(conn) (conn->upstream_proxy != NULL && conn->upstream_proxy->type == PT_HTTP)
 #else
 #  define UPSTREAM_CONFIGURED() (0)
 #  define UPSTREAM_HOST(host) (NULL)
+#  define UPSTREAM_IS_HTTP(up) (0)
 #endif
 
 /*
@@ -196,12 +199,12 @@ static int extract_url (const char *url, int default_port,
                         struct request_s *request)
 {
         char *p;
-        int len;
         int port;
 
         /* Split the URL on the slash to separate host from path */
         p = strchr (url, '/');
         if (p != NULL) {
+                int len;
                 len = p - url;
                 request->host = (char *) safemalloc (len + 1);
                 memcpy (request->host, url, len);
@@ -267,6 +270,17 @@ establish_http_connection (struct conn_s *connptr, struct request_s *request)
                                       "Connection: close\r\n",
                                       request->method, request->path,
                                       request->host, portbuff);
+        } else if (connptr->upstream_proxy &&
+                   connptr->upstream_proxy->type == PT_HTTP &&
+                   connptr->upstream_proxy->ua.authstr) {
+                return write_message (connptr->server_fd,
+                                      "%s %s HTTP/1.0\r\n"
+                                      "Host: %s%s\r\n"
+                                      "Connection: close\r\n"
+                                      "Proxy-Authorization: Basic %s\r\n",
+                                      request->method, request->path,
+                                      request->host, portbuff,
+                                      connptr->upstream_proxy->ua.authstr);
         } else {
                 return write_message (connptr->server_fd,
                                       "%s %s HTTP/1.0\r\n"
@@ -369,12 +383,18 @@ BAD_REQUEST_ERROR:
 
                 reverse_url = reverse_rewrite_url (connptr, hashofheaders, url);
 
-                if (!reverse_url) {
+                if (reverse_url != NULL) {
+                        safefree (url);
+                        url = reverse_url;
+                } else if (config.reverseonly) {
+                        log_message (LOG_ERR,
+                                     "Bad request, no mapping for '%s' found",
+                                     url);
+                        indicate_http_error (connptr, 400, "Bad Request",
+                                             "detail", "No mapping found for "
+                                             "requested url", "url", url, NULL);
                         goto fail;
                 }
-
-                safefree (url);
-                url = reverse_url;
         }
 #endif
 
@@ -847,10 +867,10 @@ process_client_headers (struct conn_s *connptr, hashmap_t hashofheaders)
         /*
          * Don't send headers if there's already an error, if the request was
          * a stats request, or if this was a CONNECT method (unless upstream
-         * proxy is in use.)
+         * http proxy is in use.)
          */
         if (connptr->server_fd == -1 || connptr->show_stats
-            || (connptr->connect_method && (connptr->upstream_proxy == NULL))) {
+            || (connptr->connect_method && ! UPSTREAM_IS_HTTP(connptr))) {
                 log_message (LOG_INFO,
                              "Not sending client headers to remote machine");
                 return 0;
@@ -1259,6 +1279,122 @@ static void relay_connection (struct conn_s *connptr)
         return;
 }
 
+static int
+connect_to_upstream_proxy(struct conn_s *connptr, struct request_s *request)
+{
+	unsigned len;
+	unsigned char buff[512]; /* won't use more than 7 + 255 */
+	unsigned short port;
+	size_t ulen, passlen;
+
+	struct hostent *host;
+	struct upstream *cur_upstream = connptr->upstream_proxy;
+
+	ulen = cur_upstream->ua.user ? strlen(cur_upstream->ua.user) : 0;
+	passlen = cur_upstream->pass ? strlen(cur_upstream->pass) : 0;
+
+
+	log_message(LOG_CONN,
+		    "Established connection to %s proxy \"%s\" using file descriptor %d.",
+		    proxy_type_name(cur_upstream->type), cur_upstream->host, connptr->server_fd);
+
+	if (cur_upstream->type == PT_SOCKS4) {
+
+		buff[0] = 4; /* socks version */
+		buff[1] = 1; /* connect command */
+		port = htons(request->port);
+		memcpy(&buff[2], &port, 2); /* dest port */
+		host = gethostbyname(request->host);
+		memcpy(&buff[4], host->h_addr_list[0], 4); /* dest ip */
+		buff[8] = 0; /* user */
+		if (9 != safe_write(connptr->server_fd, buff, 9))
+			return -1;
+		if (8 != safe_read(connptr->server_fd, buff, 8))
+			return -1;
+		if (buff[0]!=0 || buff[1]!=90)
+			return -1;
+
+	} else if (cur_upstream->type == PT_SOCKS5) {
+
+		/* init */
+		int n_methods = ulen ? 2 : 1;
+		buff[0] = 5; /* socks version */
+		buff[1] = n_methods; /* number of methods  */
+		buff[2] = 0; /* no auth method */
+		if (ulen) buff[3] = 2;  /* auth method -> username / password */
+		if (2+n_methods != safe_write(connptr->server_fd, buff, 2+n_methods))
+			return -1;
+		if (2 != safe_read(connptr->server_fd, buff, 2))
+			return -1;
+		if (buff[0] != 5 || (buff[1] != 0 && buff[1] != 2))
+			return -1;
+
+		if (buff[1] == 2) {
+			/* authentication */
+			char in[2];
+			char out[515];
+			char *cur = out;
+			size_t c;
+			*cur++ = 1;	/* version */
+			c = ulen & 0xFF;
+			*cur++ = c;
+			memcpy(cur, cur_upstream->ua.user, c);
+			cur += c;
+			c = passlen & 0xFF;
+			*cur++ = c;
+			memcpy(cur, cur_upstream->pass, c);
+			cur += c;
+
+			if((cur - out) != safe_write(connptr->server_fd, out, cur - out))
+				return -1;
+
+			if(2 != safe_read(connptr->server_fd, in, 2))
+				return -1;
+			if(in[1] != 0 || !(in[0] == 5 || in[0] == 1)) {
+				return -1;
+			}
+		}
+		/* connect */
+		buff[0] = 5; /* socks version */
+		buff[1] = 1; /* connect */
+		buff[2] = 0; /* reserved */
+		buff[3] = 3; /* domainname */
+		len=strlen(request->host);
+		if(len>255)
+			return -1;
+		buff[4] = len; /* length of domainname */
+		memcpy(&buff[5], request->host, len); /* dest ip */
+		port = htons(request->port);
+		memcpy(&buff[5+len], &port, 2); /* dest port */
+		if (7+len != safe_write(connptr->server_fd, buff, 7+len))
+			return -1;
+		if (4 != safe_read(connptr->server_fd, buff, 4))
+			return -1;
+		if (buff[0]!=5 || buff[1]!=0)
+			return -1;
+		switch(buff[3]) {
+			case 1: len=4; break; /* ip v4 */
+			case 4: len=16; break; /* ip v6 */
+			case 3: /* domainname */
+				if (1 != safe_read(connptr->server_fd, buff, 1))
+					return -1;
+				len = buff[0]; /* max = 255 */
+				break;
+			default: return -1;
+		}
+		if (2+len != safe_read(connptr->server_fd, buff, 2+len))
+			return -1;
+	} else {
+		return -1;
+	}
+
+	if (connptr->connect_method)
+		return 0;
+
+	return establish_http_connection(connptr, request);
+}
+
+
 /*
  * Establish a connection to the upstream proxy server.
  */
@@ -1302,6 +1438,9 @@ connect_to_upstream (struct conn_s *connptr, struct request_s *request)
                 return -1;
         }
 
+	if (cur_upstream->type != PT_HTTP)
+		return connect_to_upstream_proxy(connptr, request);
+
         log_message (LOG_CONN,
                      "Established connection to upstream proxy \"%s\" "
                      "using file descriptor %d.",
@@ -1469,6 +1608,38 @@ void handle_connection (int fd)
                 goto fail;
         }
 
+        if (config.basicauth_list != NULL) {
+                ssize_t len;
+                char *authstring;
+                int failure = 1;
+                len = hashmap_entry_by_key (hashofheaders, "proxy-authorization",
+                                            (void **) &authstring);
+
+                if (len == 0) {
+                        update_stats (STAT_DENIED);
+                        indicate_http_error (connptr, 407, "Proxy Authentication Required",
+                                             "detail",
+                                             "This proxy requires authentication.",
+                                             NULL);
+                        goto fail;
+                }
+                if ( /* currently only "basic" auth supported */
+                        (strncmp(authstring, "Basic ", 6) == 0 ||
+                         strncmp(authstring, "basic ", 6) == 0) &&
+                        basicauth_check (config.basicauth_list, authstring + 6) == 1)
+                                failure = 0;
+                if(failure) {
+                        update_stats (STAT_DENIED);
+                        indicate_http_error (connptr, 401, "Unauthorized",
+                                             "detail",
+                                             "The administrator of this proxy has not configured "
+                                             "it to service requests from you.",
+                                             NULL);
+                        goto fail;
+                }
+                hashmap_remove (hashofheaders, "proxy-authorization");
+        }
+
         /*
          * Add any user-specified headers (AddHeader directive) to the
          * outgoing HTTP request.
@@ -1521,7 +1692,7 @@ void handle_connection (int fd)
                 goto fail;
         }
 
-        if (!(connptr->connect_method && (connptr->upstream_proxy == NULL))) {
+        if (!connptr->connect_method || UPSTREAM_IS_HTTP(connptr)) {
                 if (process_server_headers (connptr) < 0) {
                         update_stats (STAT_BADCONN);
                         goto fail;
